from dotenv import load_dotenv
from .location_local_constants import LocationLocalConstants
from .point import Point
from .region_ml import RegionMl
from .util import LocationsUtil
load_dotenv()
from database_mysql_local.generic_crud import GenericCRUD  # noqa402
from logger_local.Logger import Logger  # noqa: E402
from language_remote.lang_code import LangCode  # noqa: E402
from user_context_remote.user_context import UserContext  # noqa: E402

logger = Logger.create_logger(
    object=LocationLocalConstants.OBJECT_FOR_LOGGER_CODE)
user_context = UserContext()


class Region(GenericCRUD):
    region_ml = RegionMl()

    def __init__(self):
        logger.start("start init Region")
        GenericCRUD.__init__(
            self,
            default_schema_name=LocationLocalConstants.LOCATION_SCHEMA_NAME,
            default_table_name=LocationLocalConstants.REGION_TABLE_NAME,
            default_view_table_name=LocationLocalConstants.REGION_VIEW_NAME,
            default_id_column_name=LocationLocalConstants.REGION_ID_COLUMN_NAME
        )

        logger.end("end init Region")

    def insert(
            self, coordinate: Point,
            region: str, lang_code: LangCode = None, title_approved: bool = False,
            country_id: int = None, group_id: int = None) -> int:
        logger.start("start insert Region",
                     object={'coordinate': coordinate, 'region': region,
                             'lang_code': lang_code,
                             'title_approved': title_approved,
                             'country_id': country_id, 'group_id': group_id})
        is_valid = LocationsUtil.validate_insert_args(
            name=region, lang_code=lang_code, title_approved=title_approved, coordinate=coordinate)
        if not is_valid:
            logger.end(log_message="Region was not inserted because no region name was provided")
            return None
        lang_code = lang_code or user_context.get_effective_profile_preferred_lang_code()
        region_json = {
            key: value for key, value in {
                'coordinate': coordinate,
                'country_id': country_id,
                'group_id': group_id
            }.items() if value is not None
        }

        region_id = GenericCRUD.insert(self, data_json=region_json)

        region_ml_id = self.region_ml.insert(region_id=region_id,
                                             region=region,
                                             lang_code=lang_code,
                                             title_approved=title_approved)

        logger.end("end insert region",
                   object={'region_id': region_id,
                           'region_ml_id': region_ml_id})
        return region_id

    def read(self, location_id: int):
        logger.start("start read location",
                     object={'location_id': location_id})
        result = GenericCRUD.select_one_dict_by_id(
            self,
            id_column_value=location_id,
            select_clause_value=LocationLocalConstants.REGION_TABLE_COLUMNS)
        result = LocationsUtil.extract_coordinates_and_replace_by_point(
            data_json=result)
        logger.end("end read location",
                   object={"result": result})
        return result

    @staticmethod
    def get_region_id_by_region_name(region_name: str, country_id: int = None,
                                     lang_code: LangCode = None) -> int:
        logger.start("start get_region_id_by_region_name",
                     object={'region_name': region_name,
                             'country_id': country_id})
        if region_name is None:
            logger.end(log_message="end get_region_id_by_region_name",
                       object={'region_id': None})
            return None
        LangCode.validate(lang_code)
        lang_code = lang_code or user_context.get_effective_profile_preferred_lang_code()
        where_clause = f"title='{region_name}' AND lang_code='{lang_code.value}'"
        # TODO: Shall we add country_id to region_ml_table and region_ml_view?
        # if country_id is not None:
        #     where_clause += f" AND country_id={country_id}"

        region_id_json = Region.region_ml.select_one_dict_by_where(
            select_clause_value=LocationLocalConstants.REGION_ID_COLUMN_NAME,
            where=where_clause,
            order_by="region_id DESC")
        region_id = region_id_json.get(
            LocationLocalConstants.REGION_ID_COLUMN_NAME)

        logger.end("end get_region_id_by_region_name",
                   object={'region_id': region_id})
        return region_id



from dotenv import load_dotenv
from .location_local_constants import LocationLocalConstants
from .neighborhood_ml import NeighborhoodMl
from .point import Point
from .util import LocationsUtil
load_dotenv()
from database_mysql_local.generic_crud import GenericCRUD  # noqa402
from logger_local.Logger import Logger  # noqa: E402
from language_remote.lang_code import LangCode  # noqa: E402
from user_context_remote.user_context import UserContext  # noqa: E402


logger = Logger.create_logger(
    object=LocationLocalConstants.OBJECT_FOR_LOGGER_CODE)
user_context = UserContext()


class Neighborhood(GenericCRUD):
    neighborhood_ml = NeighborhoodMl()

    def __init__(self):
        logger.start("start init Neighborhood")
        GenericCRUD.__init__(
            self,
            default_schema_name=LocationLocalConstants.LOCATION_SCHEMA_NAME,
            default_table_name=LocationLocalConstants.NEIGHBORHOOD_TABLE_NAME,
            default_view_table_name=LocationLocalConstants.NEIGHBORHOOD_VIEW_NAME,  # noqa501
            default_id_column_name=LocationLocalConstants.NEIGHBORHOOD_ID_COLUMN_NAME)  # noqa501

        logger.end("end init Neighborhood")

    def insert(
            self, coordinate: Point,
            neighborhood: str, lang_code: LangCode = None,
            title_approved: bool = False, city_id: int = None,
            group_id: int = None) -> int:
        logger.start("start insert neighborhood",
                     object={'coordinate': coordinate,
                             'neighborhood': neighborhood,
                             'lang_code': lang_code,
                             'title_approved': title_approved,
                             'city_id': city_id,
                             'group_id': group_id})
        is_valid = LocationsUtil.validate_insert_args(
            name=neighborhood, lang_code=lang_code, title_approved=title_approved, coordinate=coordinate)
        if not is_valid:
            logger.end(log_message="Neighborhood was not inserted because no neighborhood name was provided")
            return None
        lang_code = lang_code or user_context.get_effective_profile_preferred_lang_code()
        neighborhood_json = {
                key: value for key, value in {
                    'coordinate': coordinate,
                    'city_id': city_id,
                    'group_id': group_id
                }.items() if value is not None
            }
        neighborhood_id = GenericCRUD.insert(self, data_json=neighborhood_json)

        neighborhood_ml_id = self.neighborhood_ml.insert(
            neighborhood=neighborhood, neighborhood_id=neighborhood_id,
            lang_code=lang_code, title_approved=title_approved)

        logger.end("end insert neighborhood",
                   object={'neighborhood_id': neighborhood_id,
                           'neighborhood_ml_id': neighborhood_ml_id})
        return neighborhood_id

    def read(self, location_id: int):
        logger.start("start read location",
                     object={'location_id': location_id})
        result = GenericCRUD.select_one_dict_by_id(
            self,
            id_column_value=location_id,
            select_clause_value=LocationLocalConstants.NEIGHBORHOOD_TABLE_COLUMNS)  # noqa501

        result = LocationsUtil.extract_coordinates_and_replace_by_point(
            data_json=result)
        logger.end("end read location",
                   object={"result": result})
        return result

    @staticmethod
    def get_neighborhood_id_by_neighborhood_name(neighborhood_name: str, city_id: int = None,
                                                 lang_code: LangCode = None) -> int:
        # TODO should return also the city_id if known
        logger.start("start get_neighborhood_id_by_neighborhood_name",
                     object={'neighborhood_name': neighborhood_name,
                             'city_id': city_id})
        if neighborhood_name is None:
            logger.end(log_message="end get_neighborhood_id_by_neighborhood_name",
                       object={'neighborhood_id': None})
            return None
        LangCode.validate(lang_code)
        lang_code = lang_code or user_context.get_effective_profile_preferred_lang_code()
        where_clause = f"title='{neighborhood_name}' AND lang_code='{lang_code.value}'"
        # TODO: Shall we add city_id to neighborhood_ml_table and neighborhood_ml_view?
        # if city_id is not None:
        #     where_clause += f" AND city_id={city_id}"

        neighborhood_id_json = Neighborhood.neighborhood_ml.select_one_dict_by_where(  # noqa501
            select_clause_value=LocationLocalConstants.NEIGHBORHOOD_ID_COLUMN_NAME,  # noqa501
            where=where_clause,
            order_by="neighborhood_id DESC")
        neighborhood_id = neighborhood_id_json.get(
            LocationLocalConstants.NEIGHBORHOOD_ID_COLUMN_NAME)

        logger.end("end get_neighborhood_id_by_neighborhood_name",
                   object={'neighborhood_id': neighborhood_id})
        return neighborhood_id



from logger_local.LoggerComponentEnum import LoggerComponentEnum


class LocationLocalConstants:
    LOCATION_LOCAL_PYTHON_PACKAGE_COMPONENT_ID = 113
    LOCATION_LOCAL_PYTHON_PACKAGE_COMPONENT_NAME = 'location local python package'  # noqa501
    OBJECT_FOR_LOGGER_CODE = {
        'component_id': LOCATION_LOCAL_PYTHON_PACKAGE_COMPONENT_ID,
        'component_name': LOCATION_LOCAL_PYTHON_PACKAGE_COMPONENT_NAME,
        'component_category': LoggerComponentEnum.ComponentCategory.Code.value,
        'developer_email': 'tal.g@circ.zone'
    }

    OBJECT_FOR_LOGGER_TEST = {
        'component_id': LOCATION_LOCAL_PYTHON_PACKAGE_COMPONENT_ID,
        'component_name': LOCATION_LOCAL_PYTHON_PACKAGE_COMPONENT_NAME,
        'component_category': LoggerComponentEnum.ComponentCategory.Unit_Test.value,  # noqa501
        'developer_email': 'tal.g@circ.zone'
    }

    UNKNOWN_LOCATION_ID = 0
    DEFAULT_NEGIHBORHOOD_NAME = None
    DEFAULT_COUNTY_NAME = None
    DEFAULT_REGION_NAME = None
    '''Cannot be None because it would result in
    mysql.connector.errors.IntegrityError: 1048 (23000):
    Column 'state_name' cannot be null'''
    DEFAULT_STATE_NAME = 'UNKNOWN'
    DEFAULT_COUNTRY_NAME = 'UNKNOWN'  # Cannot be None, `name` is an index
    DEFAULT_ADDRESS_LOCAL_LANGUAGE = None
    DEFAULT_ADDRESS_ENGLISH = None
    DEFAULT_POSTAL_CODE = None
    DEFAULT_PLUS_CODE = None
    DEFAULT_COORDINATE = {'latitude': 0.0, 'longitude': 0.0}
    LOCATION_SCHEMA_NAME = 'location'
    LOCATION_TABLE_NAME = 'location_table'
    LOCATION_VIEW_NAME = 'location_view'
    LOCATION_ID_COLUMN_NAME = 'location_id'
    COUNTRY_TABLE_NAME = 'country_table'
    COUNTRY_VIEW_NAME = 'country_view'
    COUNTRY_ID_COLUMN_NAME = 'country_id'
    COUNTRY_ML_TABLE_NAME = 'country_ml_table'
    COUNTRY_ML_VIEW_NAME = 'country_ml_view'
    COUNTRY_ML_ID_COLUMN_NAME = 'country_ml_id'
    COUNTY_TABLE_NAME = 'county_table'
    COUNTY_VIEW_NAME = 'county_view'
    COUNTY_ID_COLUMN_NAME = 'county_id'
    COUNTY_ML_TABLE_NAME = 'county_ml_table'
    COUNTY_ML_VIEW_NAME = 'county_ml_view'
    COUNTY_ML_ID_COLUMN_NAME = 'county_ml_id'
    NEIGHBORHOOD_TABLE_NAME = 'neighborhood_table'
    NEIGHBORHOOD_VIEW_NAME = 'neighborhood_view'
    NEIGHBORHOOD_ID_COLUMN_NAME = 'neighborhood_id'
    NEIGHBORHOOD_ML_TABLE_NAME = 'neighborhood_ml_table'
    NEIGHBORHOOD_ML_VIEW_NAME = 'neighborhood_ml_view'
    NEIGHBORHOOD_ML_ID_COLUMN_NAME = 'neighborhood_ml_id'
    REGION_TABLE_NAME = 'region_table'
    REGION_VIEW_NAME = 'region_view'
    REGION_ID_COLUMN_NAME = 'region_id'
    REGION_ML_TABLE_NAME = 'region_ml_table'
    REGION_ML_VIEW_NAME = 'region_ml_view'
    REGION_ML_ID_COLUMN_NAME = 'region_ml_id'
    STATE_TABLE_NAME = 'state_table'
    STATE_VIEW_NAME = 'state_view'
    STATE_ID_COLUMN_NAME = 'state_id'
    STATE_ML_TABLE_NAME = 'state_ml_table'
    STATE_ML_VIEW_NAME = 'state_ml_view'
    STATE_ML_ID_COLUMN_NAME = 'state_ml_id'
    CITY_TABLE_NAME = 'city_table'
    CITY_VIEW_NAME = 'city_view'
    CITY_ID_COLUMN_NAME = 'city_id'
    CITY_ML_TABLE_NAME = 'city_ml_table'
    CITY_ML_VIEW_NAME = 'city_ml_view'
    CITY_ML_ID_COLUMN_NAME = 'city_ml_id'
    DEFAULT_POINT_COLUMN_NAME = 'coordinate'

    LOCATION_TABLE_COLUMNS = '''location_id, ST_X(coordinate),
                                ST_Y(coordinate), address_local_language,
                                address_english, neighborhood_id, city_id,
                                county_id, region_id, state_id, country_id,
                                postal_code, plus_code, is_approved,
                                is_community_active, created_timestamp,
                                created_user_id, updated_timestamp,
                                updated_user_id, start_timestamp,
                                end_timestamp'''

    CITY_TABLE_COLUMNS = '''city_id, ST_X(coordinate), ST_Y(coordinate),
                            name, phonecode, group_id, created_timestamp,
                            created_user_id, updated_timestamp,
                            updated_user_id, new_city_id, country_id'''

    COUNTRY_TABLE_COLUMNS = '''country_id, ST_X(coordinate), ST_Y(coordinate),
                               iso, name, nicename, iso3, numcode, phonecode,
                               group_id, created_timestamp, created_user_id,
                               updated_timestamp, updated_user_id'''

    COUNTY_TABLE_COLUMNS = '''county_id, ST_X(coordinate), ST_Y(coordinate),
                              group_id, created_timestamp, created_user_id,
                              updated_timestamp, updated_user_id'''

    NEIGHBORHOOD_TABLE_COLUMNS = '''neighborhood_id, city_id, ST_X(coordinate),
                                    ST_Y(coordinate), group_id,
                                    created_timestamp, created_user_id,
                                    updated_timestamp, updated_user_id'''

    REGION_TABLE_COLUMNS = '''region_id, country_id, ST_X(coordinate),
                              ST_Y(coordinate), group_id, created_timestamp,
                              created_user_id, updated_timestamp,
                              updated_user_id'''

    STATE_TABLE_COLUMNS = '''state_id, ST_X(coordinate), ST_Y(coordinate),
                             group_id, created_timestamp, created_user_id,
                             updated_timestamp, updated_user_id'''



from typing import Dict
from dotenv import load_dotenv
from logger_local.LoggerComponentEnum import LoggerComponentEnum

from .city import City
from .country import Country
from .county import County
from .location_local_constants import LocationLocalConstants
from .neighborhood import Neighborhood
from .point import Point
from .region import Region
from .state import State
from .util import LocationsUtil
load_dotenv()
from database_mysql_local.generic_crud import GenericCRUD   # noqa402
from logger_local.Logger import Logger  # noqa: E402
from user_context_remote.user_context import UserContext  # noqa: E402
from language_remote.lang_code import LangCode  # noqa: E402

LOCATION_LOCAL_PYTHON_PACKAGE_COMPONENT_ID = LocationLocalConstants.LOCATION_LOCAL_PYTHON_PACKAGE_COMPONENT_ID  # noqa501
LOCATION_LOCAL_PYTHON_PACKAGE_COMPONENT_NAME = LocationLocalConstants.LOCATION_LOCAL_PYTHON_PACKAGE_COMPONENT_NAME  # noqa501

object_to_insert = {
    'component_id': LOCATION_LOCAL_PYTHON_PACKAGE_COMPONENT_ID,
    'component_name': LOCATION_LOCAL_PYTHON_PACKAGE_COMPONENT_NAME,
    'component_category': LoggerComponentEnum.ComponentCategory.Code.value,
    'developer_email': 'tal.g@circ.zone'
}

logger = Logger.create_logger(object=object_to_insert)

user_context = UserContext()


# TODO Create LocationsLocal.get_country_id_by_location_id() and call it
# from importer-local-python-package


class LocationsLocal(GenericCRUD):
    def __init__(self):
        logger.start("start init LocationLocal")

        GenericCRUD.__init__(
            self,
            default_schema_name=LocationLocalConstants.LOCATION_SCHEMA_NAME,
            default_table_name=LocationLocalConstants.LOCATION_TABLE_NAME,
            default_view_table_name=LocationLocalConstants.LOCATION_VIEW_NAME,
            default_id_column_name=LocationLocalConstants.LOCATION_ID_COLUMN_NAME)  # noqa501

        logger.end("end init LocationLocal")

    @staticmethod
    # TODO Shall we change it to get_location_json or _dict? - As general standard.
    def get_location_info(neighborhood_name: str, county_name: str, region_name: str, state_name: str,
                          country_name: str, city_name: str) -> dict:
        # TODO Fix the algorithm as the current will not work in all cases. i.e. same neighborhood name in different places
        logger.start("start get location ids",
                     object={'neighborhood_name': neighborhood_name,
                             'county_name': county_name,
                             'region_name': region_name,
                             'state_name': state_name,
                             'country_name': country_name,
                             'city_name': city_name})

        neighborhood_id = Neighborhood.get_neighborhood_id_by_neighborhood_name(neighborhood_name)  # noqa501
        county_id = County.get_county_id_by_county_name_state_id(county_name)
        region_id = Region.get_region_id_by_region_name(region_name)
        state_id = State.get_state_id_by_state_name(state_name)
        country_id = Country.get_country_id_by_country_name(country_name)
        city_id = City.get_city_id_by_city_name_state_id(city_name)
        location_info = {"neighborhood_id": neighborhood_id, "county_id": county_id, "region_id": region_id,
                         "state_id": state_id, "country_id": country_id, "city_id": city_id}
        logger.end(object=location_info)
        return location_info

    def insert(
            self, data: Dict[str, any],
            lang_code: LangCode = None,
            is_approved: bool = True,
            is_test_data: bool = False,
            new_country_data: Dict[str, any] = None) -> int:

        logger.start("start insert location",
                     object={'data': data, 'lang_code': lang_code,
                             'is_approved': is_approved, 'is_test_data': is_test_data,
                             'new_country_data': new_country_data})
        LangCode.validate(lang_code)
        lang_code = lang_code or user_context.get_effective_profile_preferred_lang_code()
        # TODO location_json or location_dict as general standard.
        location_info = self._check_details_and_insert_if_not_exist(
            data.get("coordinate"),
            (data.get("neighborhood"),
             data.get("county"),
             data.get("region"),
             data.get("state"),
             data.get("country"),
             data.get("city")),
            lang_code, is_approved, new_country_data)

        location_json = {
            key: value for key, value in {
                'coordinate': data.get("coordinate"),
                'address_local_language': data.get(
                    "address_local_language"),
                'address_english': data.get("address_english"),
                'neighborhood_id': location_info.get("neighborhood_id"),
                'county_id': location_info.get("county_id"),
                'region_id': location_info.get("region_id"),
                'state_id': location_info.get("state_id"),
                'country_id': location_info.get("country_id"),
                'city_id': location_info.get("city_id"),
                'postal_code': data.get("postal_code"),
                'plus_code': data.get("plus_code"),
                'is_approved': is_approved,
                'is_test_data': is_test_data
            }.items() if value is not None
        }

        location_id = GenericCRUD.insert(self, data_json=location_json)

        logger.end("end_insert location",
                   object={'location_id': location_id})
        return location_id

    def update(self, location_id: int, data: Dict[str, any],
               lang_code: LangCode = None, is_approved: bool = True):

        logger.start("start update location",
                     object={'location_id': location_id, 'data': data,
                             'lang_code': lang_code,
                             'is_approved': is_approved})

        location_info = self._check_details_and_insert_if_not_exist(
            data.get("coordinate"),
            (data.get("neighborhood"),
             data.get("county"),
             data.get("region"),
             data.get("state"),
             data.get("country"),
             data.get("city")),
            lang_code, is_approved)

        updated_location_json = {
            key: value for key, value in {
                'coordinate': data.get('coordinate'),
                'address_local_language': data.get(
                    "address_local_language"),
                'address_english': data.get("address_english"),
                'neighborhood_id': location_info.get("neighborhood_id"),
                'county_id': location_info.get("county_id"),
                'region_id': location_info.get("region_id"),
                'state_id': location_info.get("state_id"),
                'country_id': location_info.get("country_id"),
                'city_id': location_info.get("city_id"),
                'postal_code': data.get("postal_code"),
                'plus_code': data.get("plus_code"),
                'is_approved': is_approved
            }.items() if value is not None
        }
        GenericCRUD.update_by_id(
            self,
            id_column_value=location_id,
            data_json=updated_location_json
        )

        logger.end("end update location")

    def read(self, location_id: int):
        logger.start("start read location",
                     object={'location_id': location_id})
        result = GenericCRUD.select_one_dict_by_id(
            self,
            id_column_value=location_id,
            select_clause_value=LocationLocalConstants.LOCATION_TABLE_COLUMNS)

        result = LocationsUtil.extract_coordinates_and_replace_by_point(
            data_json=result)
        logger.end("end read location",
                   object={"result": result})
        return result

    def delete(self, location_id: int):
        logger.start("start delete location by id",
                     object={'location_id': location_id})
        GenericCRUD.delete_by_id(self, id_column_value=location_id)

        logger.end("end delete location by id")

    def _check_details_and_insert_if_not_exist(
            self, coordinate: Dict[str, Point],
            location_details: tuple[str, str, str, str, str, str],
            lang_code: LangCode = None, is_approved: bool = False,
            new_country_data: Dict[str, any] = None) -> tuple[int, int, int, int, int, int]:
        logger.start("start _check_details_and_insert_if_not_exist",
                     object={'coordinate': coordinate,
                             'location_details': location_details,
                             'lang_code': lang_code,
                             'new_country_data': new_country_data})
        (neighborhood_name, county_name, region_name, state_name, country_name, city_name) = location_details  # noqa501

        location_info = self.get_location_info(neighborhood_name, county_name,
                                               region_name, state_name, country_name,
                                               city_name)
        if location_info is None:
            return None

        if location_info["neighborhood_id"] is None and neighborhood_name is not None:
            neighborhood_object = Neighborhood()
            location_info["neighborhood_id"] = neighborhood_object.insert(
                coordinate=coordinate,
                neighborhood=neighborhood_name,
                lang_code=lang_code,
                title_approved=is_approved)

        if location_info["county_id"] is None and country_name is not None:
            county_object = County()
            location_info["county_id"] = county_object.insert(coordinate=coordinate,
                                                              county=county_name,
                                                              lang_code=lang_code,
                                                              title_approved=is_approved)

        if location_info["region_id"] is None and region_name is not None:
            region_object = Region()
            location_info["region_id"] = region_object.insert(coordinate=coordinate,
                                                              region=region_name,
                                                              lang_code=lang_code,
                                                              title_approved=is_approved)

        if location_info["state_id"] is None and state_name is not None:
            state_object = State()
            location_info["state_id"] = state_object.insert(coordinate=coordinate,
                                                            state=state_name,
                                                            lang_code=lang_code,
                                                            state_name_approved=is_approved)

        if location_info["country_id"] is None and country_name is not None:
            country_object = Country()
            location_info["country_id"] = country_object.insert(
                coordinate=coordinate,
                country=country_name,
                lang_code=lang_code,
                title_approved=is_approved,
                new_country_data=new_country_data)
        if location_info["city_id"] is None and city_name is not None:
            city_object = City()
            location_info["city_id"] = city_object.insert(coordinate=coordinate,
                                                          city=city_name,
                                                          lang_code=lang_code,
                                                          title_approved=is_approved)
        logger.end("end _check_details_and_insert_if_not_exist",
                   object=location_info)
        return location_info

    def get_test_location_id(self):
        logger.start("start get_test_location_id")
        test_location_id = GenericCRUD.get_test_entity_id(
            self,
            entity_name="location",
            insert_function=self.insert,
            # TODO Can we use the const of Point(0,0)?
            insert_kwargs={"data": {"coordinate": Point(0, 0)}},
            view_name=LocationLocalConstants.LOCATION_VIEW_NAME)
        logger.end("end get_test_location_id", object={'test_location_id': test_location_id})
        return test_location_id



from .location_local_constants import LocationLocalConstants
from .point import Point
from language_remote.lang_code import LangCode


class LocationsUtil:
    @staticmethod
    def extract_coordinates_and_replace_by_point(
            data_json: dict,
            point_column_name: str = None) -> dict:
        point_column_name = point_column_name or LocationLocalConstants.DEFAULT_POINT_COLUMN_NAME  # noqa501

        # Extract longitude and latitude values
        longitude = data_json.pop(f'ST_X({point_column_name})', None)
        latitude = data_json.pop(f'ST_Y({point_column_name})', None)

        if longitude is not None and latitude is not None:
            # Create Point object
            point = Point(longitude=longitude, latitude=latitude)

            # Add 'point' key to the dictionary
            data_json['point'] = point

        return data_json

    @staticmethod
    def validate_insert_args(name: str, lang_code: LangCode, title_approved: bool, coordinate: Point) -> bool:
        LangCode.validate(lang_code)
        if not name:
            return False
        elif title_approved is not None and not isinstance(title_approved, bool):
            raise ValueError('title_approved must be an instance of bool or None')
        elif coordinate is not None and not isinstance(coordinate, Point):
            raise ValueError('coordinate must be an instance of Point or None')
        return True



from typing import Dict
import os
from dotenv import load_dotenv
from opencage.geocoder import OpenCageGeocode

from .country_ml import CountryMl
from .location_local_constants import LocationLocalConstants
from .point import Point
from .util import LocationsUtil

load_dotenv()
from database_mysql_local.generic_crud import GenericCRUD  # noqa402
from logger_local.Logger import Logger  # noqa: E402
from language_remote.lang_code import LangCode  # noqa: E402
from user_context_remote.user_context import UserContext  # noqa: E402
# TODO Please use python-sdk fuction to get the environment variable
# TODO Please change to LOCATION_OPENCAGE_KEY everywhere
api_key = os.getenv("OPENCAGE_KEY")

logger = Logger.create_logger(
    object=LocationLocalConstants.OBJECT_FOR_LOGGER_CODE)
user_context = UserContext()


# TODO Can we use GenericMl
# TODO Rename the class name to CountriesLocal
# TODO Create Country class which store one country
# TODO Please add is_test_data parameter
class Country(GenericCRUD):
    country_ml = CountryMl()

    def __init__(self):
        logger.start("start init Country")

        GenericCRUD.__init__(
            self,
            default_schema_name=LocationLocalConstants.LOCATION_SCHEMA_NAME,
            default_table_name=LocationLocalConstants.COUNTRY_TABLE_NAME,
            default_view_table_name=LocationLocalConstants.COUNTRY_VIEW_NAME,
            default_id_column_name=LocationLocalConstants.COUNTRY_ID_COLUMN_NAME)  # noqa501

        logger.end("end init Country")

    def insert(self, country: str, lang_code: LangCode = None,
               title_approved: bool = False,
               new_country_data: Dict[str, any] = None,
               coordinate: Point = None) -> int:
        logger.start("start insert country",
                     object={'coordinate': coordinate, 'country': country,
                             'lang_code': lang_code,
                             'title_approved': title_approved,
                             'new_country_data': new_country_data})
        is_valid = LocationsUtil.validate_insert_args(
            name=country, lang_code=lang_code, title_approved=title_approved, coordinate=coordinate)
        if not is_valid:
            logger.end(log_message="Country was not inserted because no country name was provided")
            return None
        lang_code = lang_code or user_context.get_effective_profile_preferred_lang_code()
        new_country_data = new_country_data or {}
        try:
            country_json = {
                key: value for key, value in {
                    'coordinate': coordinate,
                    'iso': new_country_data.get("iso"),
                    'name': country,
                    'nicename': new_country_data.get("nicename"),
                    'iso3': new_country_data.get("iso3"),
                    'numcode': new_country_data.get("numcode"),
                    'phonecode': new_country_data.get("phonecode")
                }.items() if value is not None
            }
            # TODO Can we use GenericMl
            country_id = GenericCRUD.insert(self, data_json=country_json)

        except Exception as e:
            logger.exception("error in insert country")
            logger.end()
            raise e
        try:
            country_ml_id = self.country_ml.insert(
                country=country,
                country_id=country_id,
                lang_code=lang_code,
                title_approved=title_approved)
        # TODO Replace e with exception everywhere
        except Exception as e:
            logger.exception("error in insert country")
            logger.end()
            raise e
        logger.end("end insert country",
                   object={'country_id': country_id,
                           'country_ml_id': country_ml_id})
        return country_id

    # TODO: the read function is duplicated in all classes
    def read(self, location_id: int):
        logger.start("start read location",
                     object={'location_id': location_id})
        result = GenericCRUD.select_one_dict_by_id(
            self,
            id_column_value=location_id,
            select_clause_value=LocationLocalConstants.COUNTRY_TABLE_COLUMNS)

        result = LocationsUtil.extract_coordinates_and_replace_by_point(
            data_json=result)
        logger.end("end read location",
                   object={"result": result})
        return result

    @staticmethod
    def get_country_id_by_country_name(country_name: str, lang_code: LangCode = None) -> int:
        logger.start("start get_country_id_by_country_name",
                     object={'country_name': country_name})
        if country_name is None:
            logger.end(log_message="end get_country_id_by_country_name",
                       object={'country_id': None})
            return None
        LangCode.validate(lang_code)
        lang_code = lang_code or user_context.get_effective_profile_preferred_lang_code()
        where_clause = f"title='{country_name}' AND lang_code='{lang_code.value}'"

        country_id_json = Country.country_ml.select_one_dict_by_where(
            select_clause_value=LocationLocalConstants.COUNTRY_ID_COLUMN_NAME,
            where=where_clause,
            order_by="country_id DESC")
        country_id = country_id_json.get(
            LocationLocalConstants.COUNTRY_ID_COLUMN_NAME)

        logger.end("end get_country_id_by_country_name",
                   object={'country_id': country_id})
        return country_id

    @staticmethod
    # TODO Please add paramete types and return types to all methods/functions
    def get_country_name(location):
        # Create a geocoder instance
        logger.start("start get_country_name",
                     object={'location': location})

        # Define the city or state
        # TODO Please create new class GeoCode and GeoCodeOpenCage Class which interit from GeoCode Class.
        # Please move those calls to the GeoCodeOpenCage.geocode(location).
        # TODO Please add calls to ApiManagement Indirect from GeoCode class
        # TODO Shall we call OpenCageGeocode(api_key) everytime or we can do it one time globally?
        geocoder = OpenCageGeocode(api_key)

        # Use geocoding to get the location details
        results = geocoder.geocode(location)

        if results and len(results) > 0:
            first_result = results[0]
            components = first_result['components']

            # Extract the country from components
            country_name = components.get('country', '')
            if not country_name:
                # If country is not found, check for country_code
                # as an alternative
                country_name = components.get('country_code', '')
        else:
            country_name = None
            logger.error("country didnt  found for %s." % location)
        logger.end("end get_country_name",
                   object={'country_name': country_name})
        return country_name
